/**
 * This file is part of logisim-evolution.
 *
 * Logisim-evolution is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * Logisim-evolution is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with logisim-evolution. If not, see <http://www.gnu.org/licenses/>.
 *
 * Original code by Carl Burch (http://www.cburch.com), 2011.
 * Subsequent modifications by:
 *   + College of the Holy Cross
 *     http://www.holycross.edu
 *   + Haute École Spécialisée Bernoise/Berner Fachhochschule
 *     http://www.bfh.ch
 *   + Haute École du paysage, d'ingénierie et d'architecture de Genève
 *     http://hepia.hesge.ch/
 *   + Haute École d'Ingénierie et de Gestion du Canton de Vaud
 *     http://www.heig-vd.ch/
 */

package com.cburch.logisim.soc.nios2;

import static com.cburch.logisim.soc.Strings.S;

import java.util.ArrayList;

import com.cburch.logisim.circuit.CircuitState;
import com.cburch.logisim.soc.data.SocSupport;
import com.cburch.logisim.soc.util.AssemblerAsmInstruction;
import com.cburch.logisim.soc.util.AssemblerExecutionInterface;
import com.cburch.logisim.soc.util.AssemblerToken;

public class Nios2OtherControlInstructions implements AssemblerExecutionInterface {

  private static final int INSTR_TRAP = 0;
  private static final int INSTR_ERET = 1;
  private static final int INSTR_BREAK = 2;
  private static final int INSTR_BRET = 3;
  private static final int INSTR_RDCTL = 4;
  private static final int INSTR_WRCTL = 5;
  private static final int INSTR_FLUSHD = 6;
  private static final int INSTR_FLUSHDA = 7;
  private static final int INSTR_FLUSHI = 8;
  private static final int INSTR_INITD = 9;
  private static final int INSTR_INITDA = 10;
  private static final int INSTR_INITI = 11;
  private static final int INSTR_FLUSHP = 12;
  private static final int INSTR_SYNC = 13;
  
  private static final int SIGN_EXTEND = 0x100;
  private final static String[] AsmOpcodes = {"TRAP", "ERET", "BREAK", "BRET",
          "RDCTL", "WRCTL", "FLUSHD", "FLUSHDA", "FLUSHI", "INITD", "INITDA", "INITI",
          "FLUSHP", "SYNC" };
  private final static Integer[] AsmOpcs = {0x3A, 0x3A, 0x3A, 0x3A,
          0x3A, 0x3A, 0x3B, 0x1B, 0x3A, 0x33, 0x13, 0x3A,
          0x3A, 0x3A };
  private final static Integer[] AsmOpxs = {0x2D, 0x01, 0x34, 0x09,
          0x26, 0x2E, SIGN_EXTEND, SIGN_EXTEND, 0x0C, SIGN_EXTEND, SIGN_EXTEND, 0x29,
          0x04, 0x36 };

  private ArrayList<String> Opcodes = new ArrayList<String>();
  private ArrayList<Integer> OpcCodes = new ArrayList<Integer>(); 
  private ArrayList<Integer> OpxCodes = new ArrayList<Integer>(); 

  private int instruction;
  private boolean valid;
  private boolean jumped;
  private int operation;
  private int sourceA;
  private int immediate;
  
  public Nios2OtherControlInstructions() {
    for (int i = 0 ; i < AsmOpcodes.length ; i++) {
      Opcodes.add(AsmOpcodes[i].toLowerCase());
      OpcCodes.add(AsmOpcs[i]);
      OpxCodes.add(AsmOpxs[i]);
    }
  }
  
  public boolean execute(Object processorState, CircuitState circuitState) {
    jumped = false;
    if (!valid) return false;
    Nios2State.ProcessorState cpuState = (Nios2State.ProcessorState) processorState;
    long pc = SocSupport.convUnsignedInt(cpuState.getProgramCounter());
    long nextPc = pc+4;
    switch (operation) {
      case INSTR_TRAP : cpuState.writeRegister(29, SocSupport.convUnsignedLong(nextPc));
                        cpuState.interrupt();
                        jumped = true;
                        break;
      case INSTR_ERET : cpuState.endofInterrupt();
                        jumped = true;
                        break;
      case INSTR_BREAK: cpuState.breakReq();
                        jumped = true;
                        break;
      case INSTR_BRET : cpuState.breakRet();
                        jumped = true;
                        break;
      case INSTR_RDCTL: cpuState.writeRegister(sourceA, cpuState.getControlRegister(immediate));
                        break;
      case INSTR_WRCTL: cpuState.setControlRegister(immediate, cpuState.getRegisterValue(sourceA));
                        break;
      default         : /* nothing to do in simulation, these are HW dependent operations */
                        break;
    }
    return true;
  }

  public String getAsmInstruction() {
    if (!valid) return null;
    StringBuffer s = new StringBuffer();
    s.append(Opcodes.get(operation));
    while (s.length() < Nios2Support.ASM_FIELD_SIZE) s.append(" ");
    switch (operation) {
      case INSTR_BREAK   : 
      case INSTR_TRAP    : if (immediate != 0) s.append(Integer.toString(immediate));
                           break;
      case INSTR_RDCTL   : s.append(Nios2State.registerABINames[sourceA]+",ctl"+immediate);
                           break;
      case INSTR_WRCTL   : s.append("ctl"+immediate+","+Nios2State.registerABINames[sourceA]);
                           break;
      case INSTR_INITD   :
      case INSTR_INITDA  :
      case INSTR_FLUSHDA :
      case INSTR_FLUSHD  : int imm = ((immediate<<16)>>16);
    	                   s.append(imm+"("+Nios2State.registerABINames[sourceA]+")");
                           break;
      case INSTR_INITI   :
      case INSTR_FLUSHI  : s.append(Nios2State.registerABINames[sourceA]);
                           break;
    }
    return s.toString();
  }

  public int getBinInstruction() { return instruction; }

  public boolean setAsmInstruction(AssemblerAsmInstruction instr) {
    valid = false;
    if (!Opcodes.contains(instr.getOpcode().toLowerCase())) return false;
    operation = Opcodes.indexOf(instr.getOpcode().toLowerCase());
    valid = true;
    int first = -1;
    switch (operation) {
      case INSTR_BREAK   :
      case INSTR_TRAP    : if (instr.getNrOfParameters() == 0) {
    	                     immediate = 0;
    	                     sourceA = 0;
                           } else if (instr.getNrOfParameters() == 1) {
                             AssemblerToken[] imm = instr.getParameter(0);
                             if (imm.length != 1 || !imm[0].isNumber()) {
                               valid = false;
                               instr.setError(imm[0], S.getter("AssemblerExpectedImmediateValue"));
                             }
                             immediate = imm[0].getNumberValue();
                             if (immediate > 0x1F || immediate < 0) {
                               valid = false;
                               instr.setError(imm[0], S.getter("AssemblerImmediateOutOfRange"));
                             }
                           } else {
                             valid = false;
                             instr.setError(instr.getInstruction(), S.getter("AssemblerExpectedZeroOrOneArgument"));
                           }
                           break;
      case INSTR_WRCTL   : first = 1; 
      case INSTR_RDCTL   : if (instr.getNrOfParameters() != 2) {
    	                     valid = false;
    	                     instr.setError(instr.getInstruction(), S.getter("AssemblerExpectedTwoArguments"));
    	                     return true;
                           }
                           if (first < 0) first = 0;
                           valid &= Nios2Support.isCorrectRegister(instr, first);
                           sourceA = Nios2Support.getRegisterIndex(instr, first);
                           first += 1;
                           first &= 1;
                           AssemblerToken[] ctl = instr.getParameter(first);
                           if (ctl.length != 1 || ctl[0].getType() != Nios2Assembler.CONTROL_REGISTER) {
                             valid = false;
                             instr.setError(ctl[0], S.getter("Nios2ExpectedControlRegister"));
                           }
                           immediate = Nios2State.getRegisterIndex(ctl[0].getValue());
                           if (immediate < 0 || immediate > 31) {
                             valid = false;
                             instr.setError(ctl[0], S.getter("AssemblerUnknownRegister"));
                           }
                           break;
      case INSTR_INITI   :
      case INSTR_FLUSHI  : if (instr.getNrOfParameters() != 1) {
    	                     valid = false;
    	                     instr.setError(instr.getInstruction(), S.getter("AssemblerExpectedOneArgument"));
    	                     return true;
                           }
    	                   valid &= Nios2Support.isCorrectRegister(instr, 0);
    	                   sourceA = Nios2Support.getRegisterIndex(instr, 0);
    	                   immediate = 0;
                           break;
      case INSTR_INITD   :
      case INSTR_INITDA  :
      case INSTR_FLUSHDA :
      case INSTR_FLUSHD  : if (instr.getNrOfParameters() != 1) {
                             valid = false;
                             instr.setError(instr.getInstruction(), S.getter("AssemblerExpectedOneArgument"));
                             return true;
                           }
                           AssemblerToken[] ireg = instr.getParameter(0);
                           if (ireg.length != 2) {
                             valid = false;
                             instr.setError(ireg[0], S.getter("Nios2AssemblerExpectedImmediateIndexedRegister"));
                             return true;
                           }
                           if (!ireg[0].isNumber()) {
                             valid = false;
                             instr.setError(ireg[0], S.getter("AssemblerExpectedImmediateValue"));
                           }
                           immediate = ireg[0].getNumberValue();
                           if (immediate >= (1<<15) || immediate < -(1<<15)) {
                             valid = false;
                             instr.setError(ireg[0], S.getter("AssemblerImmediateOutOfRange"));
                           }
                           if (ireg[1].getType() != AssemblerToken.BRACKETED_REGISTER) {
                             valid = false;
                             instr.setError(ireg[1], S.getter("Nios2AssemblerExpectedBracketedRegister"));
                             return true;
                           }
                           if (Nios2State.isCustomRegister(ireg[1].getValue())) {
                             valid = false;
                             instr.setError(ireg[1], S.getter("Nios2CannotUseCustomRegister"));
                             return true;
                           }
                           if (Nios2State.isControlRegister(ireg[1].getValue())) {
                             valid = false;
                             instr.setError(ireg[1], S.getter("Nios2CannotUseControlRegister"));
                             return true;
                           }
                           sourceA = Nios2State.getRegisterIndex(ireg[1].getValue());
                           if (sourceA < 0 || sourceA > 31) {
                             valid = false;
                             instr.setError(ireg[1], S.getter("AssemblerUnknownRegister"));
                           }
                           break;
      default            : if (instr.getNrOfParameters() != 0) {
    	                     valid = false;
    	                     instr.setError(instr.getInstruction(), S.getter("AssemblerExpectedNoArguments"));
                           }
                           break;
    }
    if (valid) {
      switch (operation) {
        case INSTR_TRAP   : instruction = Nios2Support.getRTypeInstructionCode(0, 0, 0x1d, 0x2d, immediate);
                            break;
        case INSTR_ERET   : instruction = Nios2Support.getRTypeInstructionCode(0x1d, 0x1e, 0, 0x01);
                            break;
        case INSTR_BREAK  : instruction = Nios2Support.getRTypeInstructionCode(0, 0, 0x1e, 0x34, immediate);
                            break;
        case INSTR_BRET   : instruction = Nios2Support.getRTypeInstructionCode(0x1E, 0, 0x1E, 0x09);
                            break;
        case INSTR_RDCTL  : instruction = Nios2Support.getRTypeInstructionCode(0, 0, sourceA, 0x26, immediate);
                            break;
        case INSTR_WRCTL  : instruction = Nios2Support.getRTypeInstructionCode(sourceA, 0, 0, 0x2E, immediate);
                            break;
        case INSTR_FLUSHP :
        case INSTR_INITI  : 
        case INSTR_FLUSHI : instruction = Nios2Support.getRTypeInstructionCode(sourceA, 0, 0, OpxCodes.get(operation));
                            break;
        case INSTR_SYNC   : instruction = Nios2Support.getRTypeInstructionCode(0, 0, 0, 0x36);
                            break;
        case INSTR_INITDA : 
        case INSTR_INITD  : 
        case INSTR_FLUSHDA: 
        case INSTR_FLUSHD : instruction = Nios2Support.getITypeInstructionCode(sourceA, 0, immediate, OpcCodes.get(operation));
                            break;
        default           : valid = false; return false;
      }
      instr.setInstructionByteCode(instruction, 4);
    }
    return true;
  }

  public boolean setBinInstruction(int instr) {
    valid = false;
    instruction = instr;
    int opcode = Nios2Support.getOpcode(instr);
    if (opcode == 0x3A) {
      int opx = Nios2Support.getOPXCode(instr, Nios2Support.R_TYPE);
      if (!OpxCodes.contains(opx)) return false;
      valid = true;
      operation = OpxCodes.indexOf(opx);
      int ra = Nios2Support.getRegAIndex(instr, Nios2Support.R_TYPE);
      int rb = Nios2Support.getRegBIndex(instr, Nios2Support.R_TYPE);
      int rc = Nios2Support.getRegCIndex(instr, Nios2Support.R_TYPE);
      int imm5 = Nios2Support.getOPXImm(instr, Nios2Support.R_TYPE);
      switch (operation) {
        case INSTR_TRAP   : if (ra != 0 || rb != 0 || rc != 0x1D) valid = false;
                            immediate = imm5;
                            break;
        case INSTR_ERET   : if (ra != 0x1d || rb != 0x1e || rc != 0 || imm5 != 0) valid = false;
                            break;
        case INSTR_BREAK  : if (ra != 0 || rb != 0 || rc != 0x1E) valid = false;
                            immediate = imm5;
                            break;
        case INSTR_BRET   : if (ra != 0x1E || rb != 0 || rc != 0x1E || imm5 != 0) valid = false;
                            break;
        case INSTR_RDCTL  : if (ra != 0 || rb != 0) valid = false;
                            sourceA = rc;
                            immediate = imm5;
                            break;
        case INSTR_WRCTL  : if (rb != 0 || rc != 0) valid = false;
                            sourceA = ra;
                            immediate = imm5;
                            break;
        case INSTR_FLUSHP :
        case INSTR_INITI  : 
        case INSTR_FLUSHI : if (rb != 0 || rc != 0 || imm5 != 0) valid = false;
                            sourceA = ra;
                            break;
        case INSTR_SYNC   : if (ra != 0 || rb != 0 || rc != 0 || imm5 != 0) valid = false;
                            break;
        default           : valid = false;
                            break;
      }
    } else {
      if (!OpcCodes.contains(opcode)) return false;
      operation = OpcCodes.indexOf(opcode);
      valid = true;
      switch (operation) {
        case INSTR_INITDA  : 
        case INSTR_INITD   : 
        case INSTR_FLUSHDA : 
        case INSTR_FLUSHD  : if (Nios2Support.getRegBIndex(instr, Nios2Support.I_TYPE) != 0) valid = false;
                             sourceA = Nios2Support.getRegAIndex(instr, Nios2Support.I_TYPE);
                             immediate = Nios2Support.getImmediate(instr, Nios2Support.I_TYPE);
                             break;
        default            : valid = false;
                             break;
      }
    }
    return valid;
  }

  public boolean performedJump() { return jumped&&valid; }
  public boolean isValid() { return valid; }
  public String getErrorMessage() { return null; }
  public ArrayList<String> getInstructions() { return Opcodes; }

  public int getInstructionSizeInBytes(String instruction) {
    if (Opcodes.contains(instruction.toLowerCase())) return 4;
    return -1;
  }

}
